if [ -z ${CONDA_BUILD+x} ]; then
    source /root/worker/containers/00000ir6tb5/tmp/build/80754af9/pymongo_1606345099443/work/build_env_setup.sh
fi
python setup.py install --single-version-externally-managed --record=record.txt